==> create-fly-1.21.10-6.0.8-4-sources.jar: com/zurrtum/create/client/content/trains/track/TrackPlacementClient.java <==
package com.zurrtum.create.client.content.trains.track;

import com.zurrtum.create.AllItemTags;
import com.zurrtum.create.catnip.animation.LerpedFloat;
import com.zurrtum.create.catnip.animation.LerpedFloat.Chaser;
import com.zurrtum.create.catnip.data.Couple;
import com.zurrtum.create.catnip.data.Pair;
import com.zurrtum.create.catnip.math.AngleHelper;
import com.zurrtum.create.catnip.math.VecHelper;
import com.zurrtum.create.catnip.theme.Color;
import com.zurrtum.create.client.AllSpecialTextures;
import com.zurrtum.create.client.catnip.outliner.Outliner;
import com.zurrtum.create.client.content.equipment.blueprint.BlueprintOverlayRenderer;
import com.zurrtum.create.client.foundation.utility.CreateLang;
import com.zurrtum.create.content.trains.track.BezierConnection;
import com.zurrtum.create.content.trains.track.TrackBlock;
import com.zurrtum.create.content.trains.track.TrackBlockItem;
import com.zurrtum.create.content.trains.track.TrackPlacement;
import com.zurrtum.create.content.trains.track.TrackPlacement.PlacementInfo;
import com.zurrtum.create.infrastructure.packet.c2s.PlaceExtendedCurvePacket;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_239;
import net.minecraft.class_239.class_240;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_3965;
import net.minecraft.class_746;

public class TrackPlacementClient {
    static LerpedFloat animation = LerpedFloat.linear().startWithValue(0);
    static int lastLineCount = 0;

    static class_2338 hintPos;
    static int hintAngle;
    static Couple<List<class_2338>> hints;

    static int extraTipWarmup;

    public static void clientTick(class_310 mc) {
        class_746 player = mc.field_1724;
        class_1799 stack = player.method_6047();
        class_239 hitResult = mc.field_1765;
        int restoreWarmup = extraTipWarmup;
        extraTipWarmup = 0;

        if (hitResult == null)
            return;
        if (hitResult.method_17783() != class_240.field_1332)
            return;

        class_1268 hand = class_1268.field_5808;
        if (!stack.method_31573(AllItemTags.TRACKS)) {
            stack = player.method_6079();
            hand = class_1268.field_5810;
            if (!stack.method_31573(AllItemTags.TRACKS))
                return;
        }

        if (!stack.method_7958())
            return;

        TrackBlockItem blockItem = (TrackBlockItem) stack.method_7909();
        class_1937 level = player.method_73183();
        class_3965 bhr = (class_3965) hitResult;
        class_2338 pos = bhr.method_17777();
        class_2680 hitState = level.method_8320(pos);
        if (!(hitState.method_26204() instanceof TrackBlock) && !hitState.method_45474()) {
            pos = pos.method_10093(bhr.method_17780());
            hitState = blockItem.getPlacementState(new class_1838(player, hand, bhr));
            if (hitState == null)
                return;
        }

        if (!(hitState.method_26204() instanceof TrackBlock))
            return;

        extraTipWarmup = restoreWarmup;
        boolean maxTurns = mc.field_1690.field_1867.method_1434();
        PlacementInfo info = TrackPlacement.tryConnect(level, player, pos, hitState, stack, false, maxTurns);
        if (extraTipWarmup < 20)
            extraTipWarmup++;
        if (!info.valid || !TrackPlacement.hoveringMaxed && (info.end1Extent == 0 || info.end2Extent == 0))
            extraTipWarmup = 0;

        if (!player.method_68878() && (info.valid || !info.hasRequiredTracks || !info.hasRequiredPavement))
            BlueprintOverlayRenderer.displayTrackRequirements(info, player.method_6079());

        if (info.valid)
            player.method_7353(CreateLang.translateDirect("track.valid_connection").method_27692(class_124.field_1060), true);
        else if (info.message != null)
            player.method_7353(
                CreateLang.translateDirect(info.message).method_27692(info.message.equals("track.second_point") ? class_124.field_1068 : class_124.field_1061),
                true
            );

        if (bhr.method_17780() == class_2350.field_11036) {
            class_243 lookVec = player.method_5720();
            int lookAngle = (int) (22.5 + AngleHelper.deg(class_3532.method_15349(lookVec.field_1350, lookVec.field_1352)) % 360) / 8;

            if (!pos.equals(hintPos) || lookAngle != hintAngle) {
                hints = Couple.create(ArrayList::new);
                hintAngle = lookAngle;
                hintPos = pos;

                for (int xOffset = -2; xOffset <= 2; xOffset++) {
                    for (int zOffset = -2; zOffset <= 2; zOffset++) {
                        class_2338 offset = pos.method_10069(xOffset, 0, zOffset);
                        PlacementInfo adjInfo = TrackPlacement.tryConnect(level, player, offset, hitState, stack, false, maxTurns);
                        hints.get(adjInfo.valid).add(offset.method_10074());
                    }
                }
            }

            if (hints != null && !hints.either(Collection::isEmpty)) {
                Outliner.getInstance().showCluster("track_valid", hints.getFirst()).withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
                    .colored(0x95CD41).lineWidth(0);
                Outliner.getInstance().showCluster("track_invalid", hints.getSecond()).withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
                    .colored(0xEA5C2B).lineWidth(0);
            }
        }

        animation.chase(info.valid ? 1 : 0, 0.25, Chaser.EXP);
        animation.tickChaser();

        if (!info.valid) {
            info.end1Extent = 0;
            info.end2Extent = 0;
        }

        int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
        class_243 up = new class_243(0, 4 / 16f, 0);

        {
            class_243 v1 = info.end1;
            class_243 a1 = info.axis1.method_1029();
            class_243 n1 = info.normal1.method_1036(a1).method_1021(15 / 16f);
            class_243 o1 = a1.method_1021(0.125f);
            class_243 ex1 = a1.method_1021((info.end1Extent - (info.curve == null && info.end1Extent > 0 ? 2 : 0)) * info.axis1.method_1033());
            line(1, v1.method_1019(n1).method_1019(up), o1, ex1);
            line(2, v1.method_1020(n1).method_1019(up), o1, ex1);

            class_243 v2 = info.end2;
            class_243 a2 = info.axis2.method_1029();
            class_243 n2 = info.normal2.method_1036(a2).method_1021(15 / 16f);
            class_243 o2 = a2.method_1021(0.125f);
            class_243 ex2 = a2.method_1021(info.end2Extent * info.axis2.method_1033());
            line(3, v2.method_1019(n2).method_1019(up), o2, ex2);
            line(4, v2.method_1020(n2).method_1019(up), o2, ex2);
        }

        BezierConnection bc = info.curve;
        if (bc == null)
            return;

        class_243 previous1 = null;
        class_243 previous2 = null;
        int railcolor = color;
        int segCount = bc.getSegmentCount();

        float s = animation.getValue() * 7 / 8f + 1 / 8f;
        float lw = animation.getValue() * 1 / 16f + 1 / 16f;
        class_243 end1 = bc.starts.getFirst();
        class_243 end2 = bc.starts.getSecond();
        class_243 finish1 = end1.method_1019(bc.axes.getFirst().method_1021(bc.getHandleLength()));
        class_243 finish2 = end2.method_1019(bc.axes.getSecond().method_1021(bc.getHandleLength()));
        String key = "curve";

        for (int i = 0; i <= segCount; i++) {
            float t = i / (float) segCount;
            class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
            class_243 derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, t).method_1029();
            class_243 normal = bc.getNormal(t).method_1036(derivative).method_1021(15 / 16f);
            class_243 rail1 = result.method_1019(normal).method_1019(up);
            class_243 rail2 = result.method_1020(normal).method_1019(up);

            if (previous1 != null) {
                class_243 middle1 = rail1.method_1019(previous1).method_1021(0.5f);
                class_243 middle2 = rail2.method_1019(previous2).method_1021(0.5f);
                Outliner.getInstance().showLine(Pair.of(key, i * 2), VecHelper.lerp(s, middle1, previous1), VecHelper.lerp(s, middle1, rail1))
                    .colored(railcolor).disableLineNormals().lineWidth(lw);
                Outliner.getInstance().showLine(Pair.of(key, i * 2 + 1), VecHelper.lerp(s, middle2, previous2), VecHelper.lerp(s, middle2, rail2))
                    .colored(railcolor).disableLineNormals().lineWidth(lw);
            }

            previous1 = rail1;
            previous2 = rail2;
        }

        for (int i = segCount + 1; i <= lastLineCount; i++) {
            Outliner.getInstance().remove(Pair.of(key, i * 2));
            Outliner.getInstance().remove(Pair.of(key, i * 2 + 1));
        }

        lastLineCount = segCount;
    }

    private static void line(int id, class_243 v1, class_243 o1, class_243 ex) {
        int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
        Outliner.getInstance().showLine(Pair.of("start", id), v1.method_1020(o1), v1.method_1019(ex)).lineWidth(1 / 8f).disableLineNormals().colored(color);
    }

    public static class_1269 sendExtenderPacket(class_746 player, class_1268 hand) {
        class_1799 stack = player.method_5998(hand);
        if (!stack.method_31573(AllItemTags.TRACKS))
            return null;
        if (class_310.method_1551().field_1690.field_1867.method_1434()) {
            player.field_3944.method_52787(new PlaceExtendedCurvePacket(hand == class_1268.field_5808, true));
            return class_1269.field_5812;
        }
        return null;
    }
}

==> create-fly-1.21.11-6.0.8-2-sources.jar: com/zurrtum/create/client/content/trains/track/TrackPlacementClient.java <==
package com.zurrtum.create.client.content.trains.track;

import com.zurrtum.create.AllItemTags;
import com.zurrtum.create.catnip.animation.LerpedFloat;
import com.zurrtum.create.catnip.animation.LerpedFloat.Chaser;
import com.zurrtum.create.catnip.data.Couple;
import com.zurrtum.create.catnip.data.Pair;
import com.zurrtum.create.catnip.math.AngleHelper;
import com.zurrtum.create.catnip.math.VecHelper;
import com.zurrtum.create.catnip.theme.Color;
import com.zurrtum.create.client.AllSpecialTextures;
import com.zurrtum.create.client.catnip.outliner.Outliner;
import com.zurrtum.create.client.content.equipment.blueprint.BlueprintOverlayRenderer;
import com.zurrtum.create.client.foundation.utility.CreateLang;
import com.zurrtum.create.content.trains.track.BezierConnection;
import com.zurrtum.create.content.trains.track.TrackBlock;
import com.zurrtum.create.content.trains.track.TrackBlockItem;
import com.zurrtum.create.content.trains.track.TrackPlacement;
import com.zurrtum.create.content.trains.track.TrackPlacement.PlacementInfo;
import com.zurrtum.create.infrastructure.packet.c2s.PlaceExtendedCurvePacket;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_239;
import net.minecraft.class_239.class_240;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_3965;
import net.minecraft.class_746;

public class TrackPlacementClient {
    static LerpedFloat animation = LerpedFloat.linear().startWithValue(0);
    static int lastLineCount = 0;

    static class_2338 hintPos;
    static int hintAngle;
    static Couple<List<class_2338>> hints;

    static int extraTipWarmup;

    public static void clientTick(class_310 mc) {
        class_746 player = mc.field_1724;
        class_1799 stack = player.method_6047();
        class_239 hitResult = mc.field_1765;
        int restoreWarmup = extraTipWarmup;
        extraTipWarmup = 0;

        if (hitResult == null)
            return;
        if (hitResult.method_17783() != class_240.field_1332)
            return;

        class_1268 hand = class_1268.field_5808;
        if (!stack.method_31573(AllItemTags.TRACKS)) {
            stack = player.method_6079();
            hand = class_1268.field_5810;
            if (!stack.method_31573(AllItemTags.TRACKS))
                return;
        }

        if (!stack.method_7958())
            return;

        TrackBlockItem blockItem = (TrackBlockItem) stack.method_7909();
        class_1937 level = player.method_73183();
        class_3965 bhr = (class_3965) hitResult;
        class_2338 pos = bhr.method_17777();
        class_2680 hitState = level.method_8320(pos);
        if (!(hitState.method_26204() instanceof TrackBlock) && !hitState.method_45474()) {
            pos = pos.method_10093(bhr.method_17780());
            hitState = blockItem.getPlacementState(new class_1838(player, hand, bhr));
            if (hitState == null)
                return;
        }

        if (!(hitState.method_26204() instanceof TrackBlock))
            return;

        extraTipWarmup = restoreWarmup;
        boolean maxTurns = mc.field_1690.field_1867.method_1434();
        PlacementInfo info = TrackPlacement.tryConnect(level, player, pos, hitState, stack, false, maxTurns);
        if (extraTipWarmup < 20)
            extraTipWarmup++;
        if (!info.valid || !TrackPlacement.hoveringMaxed && (info.end1Extent == 0 || info.end2Extent == 0))
            extraTipWarmup = 0;

        if (!player.method_68878() && (info.valid || !info.hasRequiredTracks || !info.hasRequiredPavement))
            BlueprintOverlayRenderer.displayTrackRequirements(info, player.method_6079());

        if (info.valid)
            player.method_7353(CreateLang.translateDirect("track.valid_connection").method_27692(class_124.field_1060), true);
        else if (info.message != null)
            player.method_7353(
                CreateLang.translateDirect(info.message)
                    .method_27692(info.message.equals("track.second_point") ? class_124.field_1068 : class_124.field_1061), true
            );

        if (bhr.method_17780() == class_2350.field_11036) {
            class_243 lookVec = player.method_5720();
            int lookAngle = (int) (22.5 + AngleHelper.deg(class_3532.method_15349(lookVec.field_1350, lookVec.field_1352)) % 360) / 8;

            if (!pos.equals(hintPos) || lookAngle != hintAngle) {
                hints = Couple.create(ArrayList::new);
                hintAngle = lookAngle;
                hintPos = pos;

                for (int xOffset = -2; xOffset <= 2; xOffset++) {
                    for (int zOffset = -2; zOffset <= 2; zOffset++) {
                        class_2338 offset = pos.method_10069(xOffset, 0, zOffset);
                        PlacementInfo adjInfo = TrackPlacement.tryConnect(level, player, offset, hitState, stack, false, maxTurns);
                        hints.get(adjInfo.valid).add(offset.method_10074());
                    }
                }
            }

            if (hints != null && !hints.either(Collection::isEmpty)) {
                Outliner.getInstance().showCluster("track_valid", hints.getFirst()).withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
                    .colored(0x95CD41).lineWidth(0);
                Outliner.getInstance().showCluster("track_invalid", hints.getSecond()).withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
                    .colored(0xEA5C2B).lineWidth(0);
            }
        }

        animation.chase(info.valid ? 1 : 0, 0.25, Chaser.EXP);
        animation.tickChaser();

        if (!info.valid) {
            info.end1Extent = 0;
            info.end2Extent = 0;
        }

        int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
        class_243 up = new class_243(0, 4 / 16f, 0);

        {
            class_243 v1 = info.end1;
            class_243 a1 = info.axis1.method_1029();
            class_243 n1 = info.normal1.method_1036(a1).method_1021(15 / 16f);
            class_243 o1 = a1.method_1021(0.125f);
            class_243 ex1 = a1.method_1021((info.end1Extent - (info.curve == null && info.end1Extent > 0 ? 2 : 0)) * info.axis1.method_1033());
            line(1, v1.method_1019(n1).method_1019(up), o1, ex1);
            line(2, v1.method_1020(n1).method_1019(up), o1, ex1);

            class_243 v2 = info.end2;
            class_243 a2 = info.axis2.method_1029();
            class_243 n2 = info.normal2.method_1036(a2).method_1021(15 / 16f);
            class_243 o2 = a2.method_1021(0.125f);
            class_243 ex2 = a2.method_1021(info.end2Extent * info.axis2.method_1033());
            line(3, v2.method_1019(n2).method_1019(up), o2, ex2);
            line(4, v2.method_1020(n2).method_1019(up), o2, ex2);
        }

        BezierConnection bc = info.curve;
        if (bc == null)
            return;

        class_243 previous1 = null;
        class_243 previous2 = null;
        int railcolor = color;
        int segCount = bc.getSegmentCount();

        float s = animation.getValue() * 7 / 8f + 1 / 8f;
        float lw = animation.getValue() * 1 / 16f + 1 / 16f;
        class_243 end1 = bc.starts.getFirst();
        class_243 end2 = bc.starts.getSecond();
        class_243 finish1 = end1.method_1019(bc.axes.getFirst().method_1021(bc.getHandleLength()));
        class_243 finish2 = end2.method_1019(bc.axes.getSecond().method_1021(bc.getHandleLength()));
        String key = "curve";

        for (int i = 0; i <= segCount; i++) {
            float t = i / (float) segCount;
            class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
            class_243 derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, t).method_1029();
            class_243 normal = bc.getNormal(t).method_1036(derivative).method_1021(15 / 16f);
            class_243 rail1 = result.method_1019(normal).method_1019(up);
            class_243 rail2 = result.method_1020(normal).method_1019(up);

            if (previous1 != null) {
                class_243 middle1 = rail1.method_1019(previous1).method_1021(0.5f);
                class_243 middle2 = rail2.method_1019(previous2).method_1021(0.5f);
                Outliner.getInstance().showLine(Pair.of(key, i * 2), VecHelper.lerp(s, middle1, previous1), VecHelper.lerp(s, middle1, rail1))
                    .colored(railcolor).disableLineNormals().lineWidth(lw);
                Outliner.getInstance().showLine(Pair.of(key, i * 2 + 1), VecHelper.lerp(s, middle2, previous2), VecHelper.lerp(s, middle2, rail2))
                    .colored(railcolor).disableLineNormals().lineWidth(lw);
            }

            previous1 = rail1;
            previous2 = rail2;
        }

        for (int i = segCount + 1; i <= lastLineCount; i++) {
            Outliner.getInstance().remove(Pair.of(key, i * 2));
            Outliner.getInstance().remove(Pair.of(key, i * 2 + 1));
        }

        lastLineCount = segCount;
    }

    private static void line(int id, class_243 v1, class_243 o1, class_243 ex) {
        int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
        Outliner.getInstance().showLine(Pair.of("start", id), v1.method_1020(o1), v1.method_1019(ex)).lineWidth(1 / 8f).disableLineNormals().colored(color);
    }

    public static class_1269 sendExtenderPacket(class_746 player, class_1268 hand) {
        class_1799 stack = player.method_5998(hand);
        if (!stack.method_31573(AllItemTags.TRACKS))
            return null;
        if (class_310.method_1551().field_1690.field_1867.method_1434()) {
            player.field_3944.method_52787(new PlaceExtendedCurvePacket(hand == class_1268.field_5808, true));
            return class_1269.field_5812;
        }
        return null;
    }
}
